/*
	privss - A private stream searching toolkit.

	Original version copyright (C) 2006 SRI International.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
	General Public License for more details.

  Later changes copyright John Bethencourt, also released under the
  GPLv2 or later.
*/

#include <alloca.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif
#include <gmp.h>
#include <iml.h>

#include <stdio.h>
#include <stdint.h>
#include <glib.h>
#include <paillier.h>
#include "privss.h"

#include <stdlib.h>
#include "linear.h"
#include "pseudorand.h"

int saved_stdout_fd = -1;

void
ditch_stdout()
{
	int new_fd;

	fflush(stdout);
  new_fd = open("/dev/null", O_WRONLY, 0666);
  saved_stdout_fd = dup(fileno(stdout));
  dup2(new_fd, fileno(stdout));
  close(new_fd);
}

void
unditch_stdout()
{
	fflush(stdout);
  dup2(saved_stdout_fd, fileno(stdout));
  close(saved_stdout_fd);
  saved_stdout_fd = -1;
}

double
get_time()
{
	static struct timeval last_tv, tv;
	static int first = 1;
	static double res = 0;

	if( first )
	{
		gettimeofday(&last_tv, 0);
		first = 0;
		return 0;
	}
	else
	{
		gettimeofday(&tv, 0);
		res += tv.tv_sec - last_tv.tv_sec;
		res += (tv.tv_usec - last_tv.tv_usec) / 1000000.0;
		last_tv = tv;
		return res;
	}
}

double
testiml( int n, int sparseness, uint32_t s )
{
	int i, j;
	double t0, t1;

	long int* A;
	mpz_t*    I;
	mpz_t*    N;
	mpz_t     D;

	/* set up A */
	A = (long int*) alloca(sizeof(long int) * n * n);
	if( !A )
		die("out of mem\n");
	for( i = 0; i < n * n; i++ )
		A[i] = !(pseudo_rand2(i, s) % sparseness);
	
	/* set up I */
	I = (mpz_t*) alloca(sizeof(mpz_t) * n * n);
	if( !I )
		die("out of mem\n");
	for( i = 0; i < n; i++ )
		for( j = 0; j < n; j++ )
			mpz_init_set_ui(I[i * n + j], i == j);

	/* set up space for results */
	N = (mpz_t*) alloca(sizeof(mpz_t) * n * n);
	if( !N )
		die("out of mem\n");
	for( i = 0; i < n * n; i++ )
		mpz_init(N[i]);
	mpz_init(D);

 	ditch_stdout();
	t0 = get_time();
	nonsingSolvMM(RightSolu, n, n, A, I, N, D);
	t1 = get_time();
 	unditch_stdout();

	for( i = 0; i < n * n; i++ )
		mpz_clear(I[i]);
	for( i = 0; i < n * n; i++ )
		mpz_clear(N[i]);
	mpz_clear(D);

	return t1 - t0;
}

double
testourold( int n, int sparseness, uint32_t s, mpz_t modulus )
{
	int i;
	double t0, t1;

	mpz_t* A;
	int* P;

	/* set up A and P */
	A = alloc_mpz_array(n * n);
	if( !A )
		die("out of mem\n");
	for( i = 0; i < n * n; i++ )
		mpz_set_ui(A[i], !(pseudo_rand2(i, s) % sparseness));
	P = alloca(sizeof(int) * n);

	/* solve */
	t0 = get_time();
	if( !lup_decomp(A, P, n, modulus) )
		die("singular A\n");
	t1 = get_time();

	/* clean up */
	free_mpz_array(A, n * n);

	return t1 - t0;
}

double
testour( int n, int sparseness, uint32_t s, mpz_t modulus )
{
	int i;
	double t0, t1;

  uint_fast8_t* A;
	mpz_t* x;
	mpz_t* b;

	/* set up A, x, and b */
  A = alloca(sizeof(uint_fast8_t) * n * n);
	for( i = 0; i < n * n; i++ )
		A[i] = !(pseudo_rand2(i, s) % sparseness);
  x = alloc_mpz_array(n);
	for( i = 0; i < n; i++ )
    mpz_set_ui(x[i], i);
  b = alloc_mpz_array(n);

	/* solve */
	t0 = get_time();
	if( !solve_system(A, x, b, n, modulus) )
		die("singular A\n");
	t1 = get_time();

	/* clean up */
  free_mpz_array(x, n);
  free_mpz_array(b, n);

	return t1 - t0;
}

void
testlen( int n, int sparseness, int trials, mpz_t modulus )
{
	int i;

	double imltime, ourtime;

	imltime = 0;
	ourtime = 0;
	for( i = 0; i < trials; i++ )
	{
 		imltime += testiml(n, sparseness, n * trials + i);
		ourtime += testour(n, sparseness, n * trials + i, modulus);
	}
	imltime /= trials;
	ourtime /= trials;

 	printf("time to solve %dx%d: %.1fs (iml) %.1fs (us)\n", n, n, imltime, ourtime);
	fflush(stdout);
}

int
main( int argc, char** argv )
{
	int n;
	int trials;
	paillier_pubkey_t* pub;
	paillier_prvkey_t* prv;

	pseudo_rand_init();

	paillier_keygen(1024, &pub, &prv, paillier_get_rand_devurandom);

	trials = 1;

	for( n = 100; n <= 1000; n += 100 )
 		testlen(n, n / 10, trials, pub->n);
/*  		testlen(n, 2, trials, pub->n); */

	return 0;
}
