/*
	privss - A private stream searching toolkit.

	Original version copyright (C) 2006 SRI International.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
	General Public License for more details.

  Later changes copyright John Bethencourt, also released under the
  GPLv2 or later.
*/

#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include <stdio.h>
#include <stdint.h>
#include <glib.h>
#include <gmp.h>
#include <paillier.h>
#include "privss.h"

/*
	Here we read and write the queries and encrypted results as binary
	files. Although human readable files are preferable, the actual data
	of these things is a space constraint and something we need to keep
	efficient. So might as well just make it binary, also for
	simplicity.
*/

void
write_uint32( FILE* fp, uint32_t k )
{
	int i;

	for( i = 3; i >= 0; i-- )
		fputc((k & 0xff<<(i*8))>>(i*8), fp);
}

void
write_uint64( FILE* fp, uint64_t k )
{
	int i;

	for( i = 7; i >= 0; i-- )
		fputc((k & 0xff<<(i*8))>>(i*8), fp);
}

void
write_string( FILE* fp, char* s )
{
	fputs(s, fp);
	fputc(0, fp);
}

void
write_pubkey( FILE* fp, paillier_pubkey_t* pub )
{
	char* s;

	s = paillier_pubkey_to_hex(pub);
	write_string(fp, s);
	free(s);
}

void
write_prvkey( FILE* fp, paillier_prvkey_t* prv )
{
	char* s;

	s = paillier_prvkey_to_hex(prv);
	write_string(fp, s);
	free(s);
}

void
write_carray( FILE* fp, paillier_pubkey_t* pub, GPtrArray* cts )
{
	int i;
	void* buf;

	write_uint32(fp, cts->len);
	for( i = 0; i < cts->len; i++ )
	{
		buf = paillier_ciphertext_to_bytes(PAILLIER_BITS_TO_BYTES(pub->bits) * 2,
																			 cts->pdata[i]);
		fwrite(buf, 1, PAILLIER_BITS_TO_BYTES(pub->bits) * 2, fp);
		free(buf);
	}
}

uint32_t
read_uint32( FILE* fp )
{
	int i;
	uint32_t r;

	r = 0;
	for( i = 3; i >= 0; i-- )
		r |= fgetc(fp)<<(i*8);

	return r;
}

uint64_t
read_uint64( FILE* fp )
{
	int i;
	uint64_t r;

	r = 0;
	for( i = 7; i >= 0; i-- )
		r |= fgetc(fp)<<(i*8);

	return r;
}

char*
read_string( FILE* fp )
{
	GString* s;
	char* r;
	char c;

	s = g_string_sized_new(32);
	while( 1 )
	{
		c = fgetc(fp);
		if( c && c != EOF )
			g_string_append_c(s, c);
		else
			break;
	}

	r = s->str;
	g_string_free(s, 0);

	return r;
}

paillier_pubkey_t*
read_pubkey( FILE* fp )
{
	paillier_pubkey_t* pub;
	char* s;

	s = read_string(fp);
	pub = paillier_pubkey_from_hex(s);
	free(s);

	return pub;
}

paillier_prvkey_t*
read_prvkey( FILE* fp, paillier_pubkey_t* pub )
{
	paillier_prvkey_t* prv;
	char* s;

	s = read_string(fp);
	prv = paillier_prvkey_from_hex(s, pub);
	free(s);

	return prv;
}

GPtrArray*
read_carray( FILE* fp, paillier_pubkey_t* pub )
{
	GPtrArray* a;
	void* buf;
	int i;

	buf = malloc(PAILLIER_BITS_TO_BYTES(pub->bits) * 2);

	a = g_ptr_array_new();
	g_ptr_array_set_size(a, read_uint32(fp));
	for( i = 0; i < a->len; i++ )
	{
 		fread(buf, 1, PAILLIER_BITS_TO_BYTES(pub->bits) * 2, fp);
		a->pdata[i] = paillier_ciphertext_from_bytes
 			(buf, PAILLIER_BITS_TO_BYTES(pub->bits) * 2);
	}

	free(buf);

	return a;
}

void
write_query( char* file, enc_query_t* query )
{
	FILE* fp;

	if( !(fp = fopen(file, "w")) )
		die("couldn't write to file: %s\n", file);

	write_uint64(fp, query->id);
	write_pubkey(fp, query->pub);
	write_carray(fp, query->pub, query->table);
	write_uint32(fp, query->pt_bytes);
	write_uint32(fp, query->block_size);
	write_uint32(fp, query->include_keywords);
	write_uint32(fp, query->i_buf_len);
	write_uint32(fp, query->c_buf_len);
	write_uint32(fp, query->l_buf_len);
	write_uint32(fp, query->d_buf_len);
	write_uint32(fp, query->num_bf_hashes);
	write_uint32(fp, query->sfactor);

	fclose(fp);
}

enc_query_t*
read_query( char* file )
{
	FILE* fp;
	enc_query_t* query;

	if( !(fp = fopen(file, "r")) )
		die("couldn't read file: %s\n", file);

	query = (enc_query_t*) malloc(sizeof(enc_query_t));

	query->id               = read_uint64(fp);
	query->pub              = read_pubkey(fp);
	query->table            = read_carray(fp, query->pub);
	query->pt_bytes 		    = read_uint32(fp);
	query->block_size       = read_uint32(fp);
	query->include_keywords = read_uint32(fp);
	query->i_buf_len 		    = read_uint32(fp);
	query->c_buf_len 		    = read_uint32(fp);
	query->l_buf_len 		    = read_uint32(fp);
	query->d_buf_len 		    = read_uint32(fp);
	query->num_bf_hashes    = read_uint32(fp);
	query->sfactor          = read_uint32(fp);

	fclose(fp);

	return query;
}

void
write_qkey( char* file, qkey_t* qkey )
{
	FILE* fp;
	GSList* words;

	if( !(fp = fopen(file, "w")) )
		die("couldn't write to file: %s\n", file);

	write_uint64(fp, qkey->id);
	write_prvkey(fp, qkey->prv);
	write_uint32(fp, g_slist_length(qkey->words));

	words = qkey->words;
	while( words )
	{
		write_string(fp, words->data);
		words = words->next;
	}

	fclose(fp);
}

qkey_t*
read_qkey( char* file, paillier_pubkey_t* pub )
{
	FILE* fp;
	qkey_t* qkey;
	int nwords;
	int i;

	if( !(fp = fopen(file, "r")) )
		die("couldn't read file: %s\n", file);

	qkey = malloc(sizeof(qkey_t));

	qkey->id  = read_uint64(fp);
	qkey->prv = read_prvkey(fp, pub);
	nwords    = read_uint32(fp);
	
	qkey->words = 0;
	for( i = 0; i < nwords; i++ )
		qkey->words = g_slist_append(qkey->words, read_string(fp));

	fclose(fp);

	return qkey;
}

void
write_results( char* file, paillier_pubkey_t* pub, enc_results_t* res )
{
	FILE* fp;

	if( !(fp = fopen(file, "w")) )
		die("couldn't write to file: %s\n", file);

	write_uint64(fp, res->id);
	write_uint32(fp, res->num_searched);
	write_carray(fp, pub, res->i_buf);
	write_carray(fp, pub, res->c_buf);
	write_carray(fp, pub, res->l_buf);
	write_carray(fp, pub, res->d_buf);

	fclose(fp);
}

enc_results_t*
read_results( char* file, enc_query_t* query, int create_new )
{
	FILE* fp;
	enc_results_t* res;
	int can_read;
	int i;

	can_read = (fp = fopen(file, "r")) && !feof(fp);
	
	if( !can_read && !create_new )
		return 0;

	res = (enc_results_t*) malloc(sizeof(enc_results_t));

	if( can_read && (res->id = read_uint64(fp)) == query->id )
	{
		/* load current search */

		res->num_searched = read_uint32(fp);
		res->i_buf        = read_carray(fp, query->pub);
		res->c_buf        = read_carray(fp, query->pub);
		res->l_buf        = read_carray(fp, query->pub);
		res->d_buf        = read_carray(fp, query->pub);
	}
	else
	{
		/* start new search */

		res->id = query->id;
		res->num_searched = 0;
		res->i_buf = g_ptr_array_new();
		res->c_buf = g_ptr_array_new();
		res->l_buf = g_ptr_array_new();
		res->d_buf = g_ptr_array_new();

		for( i = 0; i < query->i_buf_len; i++ )
			g_ptr_array_add(res->i_buf, paillier_create_enc_zero());

		for( i = 0; i < query->c_buf_len; i++ )
			g_ptr_array_add(res->c_buf, paillier_create_enc_zero());

		for( i = 0; i < query->l_buf_len; i++ )
			g_ptr_array_add(res->l_buf, paillier_create_enc_zero());

		for( i = 0; i < query->d_buf_len * query->block_size; i++ )
			g_ptr_array_add(res->d_buf, paillier_create_enc_zero());
	}

	if( fp )
		fclose(fp);

	return res;
}

int
file_exists( char* f )
{
	struct stat s;
	return stat(f, &s) == 0;
}

void
mkdirs( char* path )
{
	char* t;

	if( !strcmp(path, "/") || strlen(path) == 0 || file_exists(path) )
		return;

	t = g_dirname(path);
	mkdirs(t);
	free(t);

	mkdir(path, 0777);
}

void
write_doc( doc_t* doc, char* dir )
{
	FILE* fp;
	char* fullpath;
	char* dirdup;
	char* t;

	dirdup = strdup(dir);
	if( dirdup[strlen(dirdup) - 1] == '/' )
		dirdup[strlen(dirdup) - 1] = 0;

	fullpath = malloc
		(sizeof(char) * (strlen(dirdup) + 1 + strlen(doc->filename) + 1));

	sprintf(fullpath, "%s/%s", dirdup, doc->filename);

	t = g_dirname(fullpath);
	mkdirs(t);
	free(t);

	if( !(fp = fopen(fullpath, "w")) )
		die("couldn't write to file: %s\n", fullpath);
	fwrite(doc->content->data, 1, doc->content->len, fp);
	fclose(fp);

	free(fullpath);
	free(dirdup);
}

GSList*
slist_strdup( GSList* l )
{
	GSList* r = 0;

	while( l )
	{
		r = g_slist_append(r, strdup(l->data));
		l = l->next;
	}

	return r;
}

doc_t*
read_doc( char* file, GSList* words )
{
	FILE* fp;
	doc_t* doc;
	uint8_t buf[128];
	int len;

	doc = (doc_t*) malloc(sizeof(doc_t));

	doc->filename = g_path_is_absolute(file) ?
		strdup(g_path_skip_root(file)) : strdup(file);
	doc->words = slist_strdup(words);
	doc->content = g_byte_array_new();

	if( !(fp = fopen(file, "r")) )
		die("couldn't read file: %s\n", file);

	while( !feof(fp) )
	{
		len = fread(buf, 1, 128, fp);
		g_byte_array_append(doc->content, buf, len);
	}

	fclose(fp);

	return doc;
}

/* paillier_plaintext_t* (2D, block-major) */
GPtrArray*
serialize_doc( enc_query_t* query, doc_t* doc, int* reallen )
{
	GPtrArray* a;
	GByteArray* b;
	uint32_t nwords;
	uint8_t byte;
	int oldlen;
	GSList* w;
	int i;
	int numpts;
	int numblocks;

	/* first serialize everything into some bytes */

	b = g_byte_array_new();

	/* filename */
	g_byte_array_append(b, doc->filename, strlen(doc->filename) + 1);

	/* number of words */
	nwords = query->include_keywords ? g_slist_length(doc->words) : 0;
	for( i = 3; i >= 0; i-- )
	{
		byte = (nwords & (0xff<<(i*8)))>>(i*8);
		g_byte_array_append(b, &byte, 1);
	}

	if( query->include_keywords )
	{
		/* each of the words */
		w = doc->words;
		while( w )
		{
			g_byte_array_append(b, w->data, strlen(w->data) + 1);
			w = w->next;
		}
	}

	/* length of content */
	for( i = 3; i >= 0; i-- )
	{
		byte = (doc->content->len & (0xff<<(i*8)))>>(i*8);
		g_byte_array_append(b, &byte, 1);
	}

	/* content */
	g_byte_array_append(b, doc->content->data,  doc->content->len);

	/* then break the bytes into plaintexts */

	numpts    = CEIL_FRAC(b->len, query->pt_bytes);
	numblocks = CEIL_FRAC(numpts, query->block_size);
	a = g_ptr_array_new();
	g_ptr_array_set_size(a, numblocks * query->block_size);
	*reallen = numpts;

	/* pad out b so it evenly divides into blocks of pts */
	oldlen = b->len;
	g_byte_array_set_size(b, a->len * query->pt_bytes);
	memset(b->data + oldlen, 0, b->len - oldlen);

	/* make each of the pts */
	for( i = 0; i < a->len; i++ )
		a->pdata[i] = paillier_plaintext_from_bytes
			(b->data + i * query->pt_bytes, query->pt_bytes);

	g_byte_array_free(b, 1);

	return a;
}

char*
next_str( uint8_t** p )
{
	char* s = (char*) *p;
	while( **p )
		(*p)++;
	(*p)++;
	return s;
}

doc_t*
deserialize_doc( enc_query_t* query,
								 /* paillier_plaintext_t* (2D, block-major) */
								 GPtrArray* pts )
{
	GByteArray* b;
	doc_t* doc;
	uint8_t* p;
	uint32_t nwords;
	uint32_t len;
	int i;

	/* first put all the pts together into one byte array */

	b = g_byte_array_new();

	for( i = 0; i < pts->len; i++ )
	{
		void* buf;
		buf = paillier_plaintext_to_bytes(query->pt_bytes, pts->pdata[i]);
		g_byte_array_append(b, buf, query->pt_bytes);
		free(buf);
	}

	/* now deserialize */

	doc = malloc(sizeof(doc_t));

	/* filename */
	p = (uint8_t*) b->data;
	doc->filename = strdup(next_str(&p));

	/* number of words */
	nwords = 0;
	for( i = 0; i < 4; i++ )
	{
		nwords <<= 8;
		nwords |= *(p++);
	}

	/* each of the words */
	doc->words = 0;
	for( i = 0; i < nwords; i++ )
		doc->words = g_slist_append(doc->words, strdup(next_str(&p)));

	/* length of the content */
	len = 0;
	for( i = 0; i < 4; i++ )
	{
		len <<= 8;
		len |= *(p++);
	}

	/* content */
	doc->content = g_byte_array_new();
	g_byte_array_append(doc->content, p, len);

	/* cleanup */
	g_byte_array_free(b, 1);

	return doc;
}

void
free_carray( GPtrArray* a /* paillier_ciphertext_t* */ )
{
	int i;

	for( i = 0; i < a->len; i++ )
		paillier_freeciphertext(a->pdata[i]);
	g_ptr_array_free(a, 0);
}

void
free_parray( GPtrArray* a /* paillier_plaintext_t* */ )
{
	int i;

	for( i = 0; i < a->len; i++ )
		paillier_freeplaintext(a->pdata[i]);
	g_ptr_array_free(a, 0);
}

void
free_query( enc_query_t* query )
{
	paillier_freepubkey(query->pub);
	free_carray(query->table);
	free(query);
}

void
free_slist_items( GSList* l )
{
	while( l )
	{
		GSList* lp = l->next;
		free(l->data);
		g_slist_free_1(l);
		l = lp;
	}
}

void
free_qkey( qkey_t* qkey )
{
	paillier_freeprvkey(qkey->prv);
	free_slist_items(qkey->words);
	free(qkey);
}

void
free_results( enc_results_t* res )
{
	free_carray(res->i_buf);
	free_carray(res->c_buf);
	free_carray(res->l_buf);
	free_carray(res->d_buf);
	free(res);
}

void
free_doc( doc_t* doc )
{
	free(doc->filename);
	free_slist_items(doc->words);
	g_byte_array_free(doc->content, 1);
	free(doc);
}

int verbosity = VERBOSITY_NORMAL;

void
set_verbosity( int level )
{
	verbosity = level;
}

void
status(char* fmt, ...)
{
	va_list args;

	if( verbosity < VERBOSITY_NORMAL )
		return;

	va_start(args, fmt);
	vfprintf(stdout, fmt, args);
	va_end(args);
}

void
status_verb(char* fmt, ...)
{
	va_list args;

	if( verbosity < VERBOSITY_LOUD )
		return;

	va_start(args, fmt);
	vfprintf(stdout, fmt, args);
	va_end(args);
}

void
die(char* fmt, ...)
{
	va_list args;
	va_start(args, fmt);
	vfprintf(stderr, fmt, args);
	va_end(args);
	exit(1);
}

char*
uint64_to_hex( uint64_t i )
{
	static char buf[17];
	int j;
	int k;
	
	for( j = 0; j < 16; j++ )
	{
		k = (i & (15<<((15-j)*4)))>>((15-j)*4);
		buf[j] = k > 9 ? k - 10 + 'a' : k + '0';
	}

	buf[16] = 0;

	return buf;
}
