/*
	privss - A private stream searching toolkit.

	Original version copyright (C) 2006 SRI International.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
	General Public License for more details.

  Later changes copyright John Bethencourt, also released under the
  GPLv2 or later.
*/

#include <stdlib.h>
#include <stdint.h>
#include <gmp.h>
#include "linear.h"

/*
void
print_matrix( mpz_t* A, int len )
{
	int i;
	int j;
	
	for( i = 0; i < len; i++ )
	{
		for( j = 0; j < len; j++ )
			fputc(mpz_sgn(A[i*len + j]) == 0 ? '0' :
						mpz_cmp_ui(A[i*len + j], 1) == 0 ? '1' :
						mpz_odd_p(A[i*len + j]) ? 'x' : 'z', stdout);
		fputc('\n', stdout);
	}
}
*/

int
solve_system( uint8_t* A, mpz_t* x, mpz_t* b, int len, mpz_t n )
{
	int i;
	mpz_t* LU; /* L and U crammed into one array (not product of L and U) */
	mpz_t* y;
	mpz_t* Pb;
	int* P;    /* represented as permutation rather than matrix */

	/* do LUP decomposition on A to get L,U,P s.t. PA = LU */
  LU = alloc_mpz_array(len * len);
	for( i = 0; i < len * len; i++ )
		mpz_set_ui(LU[i], A[i]);
	P = malloc(sizeof(int) * len);
	if( !lup_decomp(LU, P, len, n) )
	{
		free_mpz_array(LU, len * len);
		free(P);
		return 0;
	}
	
	/* do forward substitution on Ly = Pb to get y s.t. y = Ux */
	y  = alloc_mpz_array(len);
	Pb = alloc_mpz_array(len);
	for( i = 0; i < len; i++ )
		mpz_set(Pb[i], b[P[i]]);
	forward_subst(LU, y, Pb, len, n);
	free_mpz_array(Pb, len);
	free(P);

	/* do back substitution on Ux = y to get x s.t. Ax = b */
	back_subst(LU, x, y, len, n);
	free_mpz_array(y, len);
	free_mpz_array(LU, len * len);

	return 1;
}

void
apply_initial_perm( mpz_t* A, int* P, int len )
{
	int i, j;
	mpz_t* T;

	T = alloc_mpz_array(len * len);

	for( i = 0; i < len; i++ )
		for( j = 0; j < len; j++ )
			mpz_set(T[i*len + j], A[P[i]*len + j]);

	for( i = 0; i < len * len; i++ )
		mpz_set(A[i], T[i]);

	free_mpz_array(T, len * len);
}

mpz_t* cmp_state_A;
int cmp_state_len;

int
cmp_leading_zeros( const void* a, const void* b )
{
	int i;
	int na, nb;

	na = 0;
	i = 0;
	while( mpz_sgn(cmp_state_A[*((int*)a) * cmp_state_len + i]) == 0 && i++ < cmp_state_len )
		na++;

	nb = 0;
	i = 0;
	while( mpz_sgn(cmp_state_A[*((int*)b) * cmp_state_len + i]) == 0 && i++ < cmp_state_len )
		nb++;

	return na < nb ? 1 : na == nb ? 0 : -1;
}

int
cmp_total_zeros( const void* a, const void* b )
{
	int i;
	int na, nb;

	na = 0;
	nb = 0;

	for( i = 0; i < cmp_state_len; i++ )
	{
		if( mpz_sgn(cmp_state_A[*((int*)a) * cmp_state_len + i]) == 0 )
			na++;
		if( mpz_sgn(cmp_state_A[*((int*)b) * cmp_state_len + i]) == 0 )
			nb++;
	}

	return na < nb ? 1 : na == nb ? 0 : -1;
}

void
pick_initial_perm( mpz_t* A, int* P, int len )
{
	/* sort in descending (?!) order by number of zeros */
	cmp_state_A = A;
	cmp_state_len = len;
 	qsort(P, len, sizeof(int), cmp_total_zeros);

	/* rearrange the rows of A according to P */
	apply_initial_perm(A, P, len);
}

void
swap_ints( int* a, int* b )
{
	int t;

	 t = *a;
	*a = *b;
	*b =  t;
}

/*
	Implemented as described on page 752 of CLRS.
*/
int
lup_decomp( mpz_t* LU, int* P, int len, mpz_t n )
{
	int i, j, k, kp;
	mpz_t kkinv;

	mpz_init(kkinv);

	for( i = 0; i < len; i++ )
		P[i] = i;

 	pick_initial_perm(LU, P, len);

	for( k = 0; k < len; k++ )
	{
		/* pick non-zero pivot k' */
		kp = k;
		while( mpz_sgn(LU[kp*len + k]) == 0 && kp < len )
			kp++;
		if( kp == len )
		{
			mpz_clear(kkinv);
			return 0;
		}
		
		/* permute accordingly */
		if( k != kp )
		{
			swap_ints(&P[k], &P[kp]);
			for( i = 0; i < len; i++ )
				mpz_swap(LU[k*len + i], LU[kp*len + i]);
		}

		mpz_invert(kkinv, LU[k*len + k], n);

		/* fill in rest of matrix */
		for( i = k + 1; i < len; i++ )
			if( mpz_sgn(LU[i*len + k]) )
			{
				/* LU[i,k] = LU[i,k] / LU[k,k] */
				mpz_mul(LU[i*len + k], LU[i*len + k], kkinv);
				mpz_mod(LU[i*len + k], LU[i*len + k], n);
				for( j = k + 1; j < len; j++ )
					if( mpz_sgn(LU[k*len + j]) )
					{
						/* LU[i,j] -= LU[i,k] * LU[k,j] */
						mpz_submul(LU[i*len + j], LU[i*len + k], LU[k*len + j]);
						mpz_mod(LU[i*len + j], LU[i*len + j], n);
					}
			}				
	}

  mpz_clear(kkinv);

	return 1;
}

void
forward_subst( mpz_t* L, mpz_t* x, mpz_t* b, int len, mpz_t n )
{
	int i, j;

	for( i = 0; i < len; i++ )
	{
		mpz_set(x[i], b[i]);
		for( j = 0; j < i - 1; j++ )
			mpz_submul(x[i], L[i*len + j], x[j]);
	}
}

void
back_subst( mpz_t* U, mpz_t* x, mpz_t* b, int len, mpz_t n )
{
	int i, j;
	mpz_t iiinv;

	mpz_init(iiinv);

	for( i = len - 1; i >= 0; i-- )
	{
		mpz_set(x[i], b[i]);
		for( j = i + 1; j < len; j++ )
			mpz_submul(x[i], U[i*len + j], x[j]);
		mpz_invert(iiinv, U[i*len + i], n);
		mpz_mul(x[i], x[i], iiinv);
		mpz_mod(x[i], x[i], n);
	}
	
	mpz_clear(iiinv);
}

mpz_t*
alloc_mpz_array( int len )
{
	int i;
	mpz_t* r;

	r = malloc(sizeof(mpz_t) * len);
	for( i = 0; i < len; i++ )
		mpz_init(r[i]);

	return r;
}

void
free_mpz_array( mpz_t* a, int len )
{
	int i;
	
	for( i = 0; i < len; i++ )
		mpz_clear(a[i]);

	free(a);
}
