/*
	privss - A private stream searching toolkit.

	Original version copyright (C) 2006 SRI International.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
	General Public License for more details.

  Later changes copyright John Bethencourt, also released under the
  GPLv2 or later.
*/

/*
	Include stdio.h, stdint.h, glib.h, gmp.h and paillier.h before
	including this file.
*/

/*
	TODO

  - performance, compatibility cleanup

    - implement LUP decomposition ourselves and get rid of iml unless
      our implementation is much slower. running LUP by hand on a
      sparse 0,1 matrix seems to suggest there is some room for clever
      optimizations. note: iml not even avail on web anymore. new
      note: it's back.

      - ok so we implemented our own lup decomp. our impl ends up
        being slower than iml for regular 0,1 matrices with equal prob
        of 0's and 1's. but when the matrices get sparse, ours gets
        faster. for sparseness = n / 10:

        time to solve 100x100:  2.1s (iml)   0.3s (us)
				time to solve 200x200:  8.7s (iml)   2.6s (us)
				time to solve 300x300: 20.5s (iml)   8.4s (us)
				time to solve 400x400: 36.3s (iml)  20.7s (us)
				time to solve 500x500:  segv (iml)  37.8s (us)
				time to solve 600x600:  segv (iml)  77.4s (us)
				time to solve 700x700:  segv (iml) 116.3s (us)
				time to solve 800x800:  segv (iml) 175.3s (us)
				time to solve 900x900:  segv (iml) 226.3s (us)
				time to solve 1000x1000:segv (iml) 352.0s (us)

      - there is not as much room for clever optimizations as i
        thought, but some. one thing that seems to help a little is
        picking the initial permutation by sorting the rows of A by
        number of leading zeros. confusingly, it seems to help when
        they are in _descending_ order!

        time to solve 100x100:   0.3s (unopt)   0.3s (sort desc)   0.4s (sort asc)
				time to solve 200x200:   3.7s (unopt)   2.6s (sort desc)   3.4s (sort asc)
				time to solve 300x300:  10.7s (unopt)   8.4s (sort desc)  11.6s (sort asc)
				time to solve 400x400:  28.3s (unopt)  20.7s (sort desc)  31.7s (sort asc)
				time to solve 500x500:  51.2s (unopt)  37.8s (sort desc)  59.4s (sort asc)
				time to solve 600x600:  99.5s (unopt)  77.4s (sort desc) 105.4s (sort asc)
				time to solve 700x700: 137.2s (unopt) 116.3s (sort desc) 149.2s (sort asc)

      - note: descending total zeros is actually better than
        descending leading zeros (ascending total zeros is bad)

			  time to solve 100x100:   0.2s (desc tot z)
				time to solve 200x200:   1.7s (desc tot z)
				time to solve 300x300:   7.5s (desc tot z)
				time to solve 400x400:  20.6s (desc tot z)
				time to solve 500x500:  34.2s (desc tot z)
				time to solve 600x600:  68.1s (desc tot z)
				time to solve 700x700:  96.5s (desc tot z)
				time to solve 800x800: 155.8s (desc tot z)
				time to solve 900x900: 200.5s (desc tot z)

      - all these times for 1024 bit keys, smaller keys are of course
        faster.

      - so overall, here's what i think we should do: change search to
        use a sparseness of n / 10 (or maybe less, figure
        experimentally with octave). then ditch iml and use our thing.
        further improvements will probably be best gained by either
        implementing strassen's algorithm or looking around the web
        for optimizations of lup.

  - interface cleanup

    - fix command line options to qcon for specifying search
      parameters once and for all. either do low-level parameters, or
      "high-level" parameters. in the later case, specify some problem
      parameters and have it analyze the situation and give you info
      about what you can expect. for setting low-level params based on
      high-level params, use malware-search/scripts/runscena as a
      guide.

    - redesign privss-search interface somewhat to be reminiscent of
      tar.

    - include the search parameters and pub key in the enc res file so
      that recon doesn't need the original query.

  - features

    - implement public dictionaries (could be quite useful in some
      situations).

    - implement OS buffer as alternative to bf (probably less useful,
      but in some situations with public dict you may want this).
*/

#define CEIL_FRAC(a,b) ((a) % (b) ? (a) / (b) + 1 : (a) / (b))
#define PMAX(a,b) ((a) < (b) ? (b) : (a))
#define PMIN(a,b) ((a) > (b) ? (b) : (a))
#define PRIVSS_VERSION_MSG PACKAGE_NAME "%s " PACKAGE_VERSION "\n" \
"\n" \
"Original verson copyright (C) 2006 SRI International, later changes\n" \
"copyright John Bethencourt.\n" \
"\n" \
"This is free software released under the GPL, see the source for copying\n" \
"conditions. There is NO warranty; not even for MERCHANTABILITY or FITNESS\n" \
"FOR A PARTICULAR PURPOSE.\n" \
"\n" \
"Written by John Bethencourt <bethenco@cs.berkeley.edu>.\n"

/* types */

typedef struct
{
	/* query id */
	uint64_t id;

	/* pub key */
	paillier_pubkey_t* pub;

	/* enc hash table */
	GPtrArray* table; /* paillier_ciphertext_t* */

	/* search parameters */
	uint32_t pt_bytes;
	uint32_t block_size;
	uint32_t include_keywords;
	uint32_t i_buf_len;
	uint32_t c_buf_len;
	uint32_t l_buf_len;
	uint32_t d_buf_len; /* in blocks (NOT equal to d_buf->len) */
	uint32_t num_bf_hashes;
	uint32_t sfactor;
} enc_query_t;

typedef struct
{
	/* query id */
	uint64_t id;
	
	/* private key */
	paillier_prvkey_t* prv;

	/* words in query (for filtering false positives) */
	GSList* words;
} qkey_t;

typedef struct
{
	/* query id */
	uint64_t id;

	/* documents processed so far */
	uint32_t num_searched;

	/* server buffers */
	GPtrArray* i_buf; /* paillier_ciphertext_t* */
	GPtrArray* c_buf; /* paillier_ciphertext_t* */
	GPtrArray* l_buf; /* paillier_ciphertext_t* */
	GPtrArray* d_buf; /* paillier_ciphertext_t* (2D, block-major) */
} enc_results_t;

typedef struct
{
	/* file and path read from on server, as specified to privss-search */
	char*       filename;

	/* words given to privss-search on server */
	GSList*     words;

	/* what's in the file */
	GByteArray* content;
} doc_t;

/* read and write various bits of things */

void write_uint32( FILE* fp, uint32_t k );
void write_uint64( FILE* fp, uint64_t k );
void write_string( FILE* fp, char* s );
void write_pubkey( FILE* fp, paillier_pubkey_t* pub );
void write_prvkey( FILE* fp, paillier_prvkey_t* prv );
void write_carray( FILE* fp, paillier_pubkey_t* pub, GPtrArray* cts );

uint32_t           read_uint32( FILE* fp );
uint64_t           read_uint64( FILE* fp );
char*              read_string( FILE* fp );
paillier_pubkey_t* read_pubkey( FILE* fp );
paillier_prvkey_t* read_prvkey( FILE* fp, paillier_pubkey_t* pub );
GPtrArray*         read_carray( FILE* fp, paillier_pubkey_t* pub );

/* query reading and writing */

void write_query( char* file, enc_query_t* query );
enc_query_t* read_query( char* file );

/* query private key reading and writing */

void write_qkey( char* file, qkey_t* qkey );
qkey_t* read_qkey( char* file, paillier_pubkey_t* pub );

/* results reading and writing */

void write_results( char* file, paillier_pubkey_t* pub, enc_results_t* results );
enc_results_t* read_results( char* file, enc_query_t* query, int create_new );

/* document handling */

void write_doc( doc_t* doc, char* dir );
doc_t* read_doc( char* file, GSList* words );

/* paillier_plaintext_t* (2D, block-major) */
GPtrArray* serialize_doc( enc_query_t* query, doc_t* doc, int* reallen );
doc_t* deserialize_doc( enc_query_t* query,
												GPtrArray* pts /* paillier_plaintext_t* (2D, block-major) */ );

/* cleanup */

void free_carray( GPtrArray* a /* paillier_ciphertext_t* */ );
void free_parray( GPtrArray* a /* paillier_plaintext_t* */ );
void free_slist_items( GSList* l );
void free_query( enc_query_t* query );
void free_qkey( qkey_t* qkey );
void free_results( enc_results_t* res );
void free_doc( doc_t* doc );

/* message printing utilities */

#define VERBOSITY_LOUD   2
#define VERBOSITY_NORMAL 1
#define VERBOSITY_QUIET  0
void set_verbosity( int level );
void status(char* fmt, ...);
void status_verb(char* fmt, ...);
void die(char* fmt, ...);

/* misc */

char* uint64_to_hex( uint64_t i ); /* returns static buffer */
