/*
	privss - A private stream searching toolkit.

	Original version copyright (C) 2006 SRI International.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
	General Public License for more details.

  Later changes copyright John Bethencourt, also released under the
  GPLv2 or later.
*/

#include <string.h>
#include <stdlib.h>

#include <stdio.h>
#include <stdint.h>
#include <glib.h>
#include <gmp.h>
#include <paillier.h>
#include "privss.h"
#include "pseudorand.h"

char* usage =
"Usage: privss-qcon [OPTION ...] WORD [WORD ...]\n"
"\n"
"Generate an encrypted query searching for each listed WORD.\n"
"\n"
"Mandatory arguments to long options are mandatory for short options too.\n\n"
" -h, --help               print this message\n\n"
" -v, --version            print version information\n\n"
" -q, --quiet              don't print status messages\n\n"
" -b, --keybits N          generate an N-bit key pair; default is 1024\n\n"
" -o, --output-query FILE  store resulting query in FILE; default is\n"
"                          \"enc_query\"\n\n"
" -k, --output-key FILE    store generated private key in FILE; default is\n"
"                          \"prv_key\"\n\n"
" -z, --block-size N       process documents in blocks of N elements;\n"
"                          default is 16\n\n"
" -w, --no-keywords        do not save keywords with searched documents;\n"
"                          in this case it will not be possible to filter\n"
"                          false positives upon reconstruction\n\n"
" -t, --table-len N        use a hash table with N entries (larger N results\n"
"                          in a larger query, smaller N results in false\n"
"                          positives); default is 256\n\n"
" -i, --i-buf-len N        use a i-buffer with N entries; default is 315\n\n"
" -c, --c-buf-len N        use a c-buffer with N entries; default is 135\n\n"
" -l, --l-buf-len N        use a l-buffer with N entries; default is 128\n\n"
" -d, --d-buf-len N        use a d-buffer with N blocks; default is 512\n\n"
"";

/* global options / data that don't go directly in the query struct */

char*	qoutfile 	= "enc_query";
char*	koutfile 	= "prv_key";
int  	keybits  	= 1024;
int  	i_buf_len = 315;
int  	c_buf_len = 135;
int   l_buf_len = 128;
int   d_buf_len = 512;
int   no_rerand = 0;

/* parses args and fills in both global options and fields in query */

void
parse_args( int argc, char** argv, enc_query_t* query, qkey_t* qkey )
{
	int i;

	g_ptr_array_set_size(query->table, 256);
	query->block_size = 16;
	query->include_keywords = 1;
	qkey->words = 0;

	for( i = 1; i < argc; i++ )
		if(      !strcmp(argv[i], "-h") || !strcmp(argv[i], "--help") )
		{
			printf("%s", usage);
			exit(0);
		}
		else if( !strcmp(argv[i], "-v") || !strcmp(argv[i], "--version") )
		{
			printf(PRIVSS_VERSION_MSG, "-qcon");
			exit(0);
		}
		else if( !strcmp(argv[i], "-q") || !strcmp(argv[i], "--quiet") )
		{
			set_verbosity(VERBOSITY_QUIET);
		}
		else if( !strcmp(argv[i], "-b") || !strcmp(argv[i], "--keybits") )
		{
			if( ++i >= argc || sscanf(argv[i], " %d ", &keybits) < 1 )
				die(usage);
		}
		else if( !strcmp(argv[i], "-o") || !strcmp(argv[i], "--output-query") )
		{
			if( ++i >= argc )
				die(usage);
			else
				qoutfile = argv[i];
		}
		else if( !strcmp(argv[i], "-k") || !strcmp(argv[i], "--output-key") )
		{
			if( ++i >= argc )
				die(usage);
			else
				koutfile = argv[i];
		}
		else if( !strcmp(argv[i], "-z") || !strcmp(argv[i], "--block-size") )
		{
			if( ++i >= argc || sscanf(argv[i], " %d ", &(query->block_size)) < 1 )
				die(usage);
		}
		else if( !strcmp(argv[i], "-w") || !strcmp(argv[i], "--no-keywords") )
		{
			query->include_keywords = 0;
		}
		else if( !strcmp(argv[i], "-t") || !strcmp(argv[i], "--table-len") )
		{
			int newlen;
			if( ++i >= argc || sscanf(argv[i], " %d ", &newlen) < 1 || newlen <= 0 )
				die(usage);
			g_ptr_array_set_size(query->table, newlen);
		}
		else if( !strcmp(argv[i], "-i") || !strcmp(argv[i], "--i-buf-len") )
		{
			if( ++i >= argc || sscanf(argv[i], " %d ", &i_buf_len) < 1 )
				die(usage);
		}
		else if( !strcmp(argv[i], "-c") || !strcmp(argv[i], "--c-buf-len") )
		{
			if( ++i >= argc || sscanf(argv[i], " %d ", &c_buf_len) < 1 )
				die(usage);
		}
		else if( !strcmp(argv[i], "-l") || !strcmp(argv[i], "--l-buf-len") )
		{
			if( ++i >= argc || sscanf(argv[i], " %d ", &l_buf_len) < 1 )
				die(usage);
		}
		else if( !strcmp(argv[i], "-d") || !strcmp(argv[i], "--d-buf-len") )
		{
			if( ++i >= argc || sscanf(argv[i], " %d ", &d_buf_len) < 1 )
				die(usage);
		}
	  /* undocumented developer options */
		else if( !strcmp(argv[i], "--xxx-no-rerandomize") )
		{
			no_rerand = 1;
		}
		else
		{
			qkey->words = g_slist_append(qkey->words, argv[i]);
		}
	
	if( !qkey->words )
		die(usage);
}

/* figure out how to set the server's parameters to meet the requirements */

void
set_params( enc_query_t* query )
{
	double nat_log_2;

	nat_log_2 = 0.6931471806;

	/*
		Ensure that each document is smaller than the modulus. We allow
		about three bytes of extra space to allow for large cvals.
	*/

	query->pt_bytes = PAILLIER_BITS_TO_BYTES(keybits) - 3;

	query->sfactor = (keybits - 16) / 16;
	query->i_buf_len = CEIL_FRAC(i_buf_len, query->sfactor);
	query->c_buf_len = PMAX(30, c_buf_len);
	query->l_buf_len = PMAX(30, l_buf_len);
	query->d_buf_len = PMAX(30, d_buf_len);

	/*
		This is the optimum number of hash functions for a Bloom filter
		with i_buf_entries entries which will store l_buf_len values.
	*/

	query->num_bf_hashes =
		PMAX(1, query->i_buf_len * query->sfactor * nat_log_2 / l_buf_len);
}

/* make the raw hash table */

int*
create_table( GSList* words, int len )
{
	int* table;

	table = (int*) malloc(sizeof(int) * len);
	memset(table, 0, sizeof(int) * len);

	while( words )
	{
		table[pseudo_rands((char*) words->data) % len] = 1;
		words = words->next;
	}

	return table;
}

/* make the actual hash table and put it in the query */

void
build_query( int* raw_hash, enc_query_t* query )
{
	int i;
	paillier_plaintext_t* tmp;

	/* only for no rerand */
	paillier_ciphertext_t* e0;
	paillier_ciphertext_t* e1;

	e0 = 0;
	e1 = 0;
	if( no_rerand )
	{
		tmp = paillier_plaintext_from_ui(0);
 		e0  = paillier_enc(0, query->pub, tmp, paillier_get_rand_devurandom);
		paillier_freeplaintext(tmp);
		tmp = paillier_plaintext_from_ui(1);
 		e1  = paillier_enc(0, query->pub, tmp, paillier_get_rand_devurandom);
		paillier_freeplaintext(tmp);
	}

	for( i = 0; i < query->table->len; i++ )
	{
		if( no_rerand )
		{
			if( raw_hash[i] )
				query->table->pdata[i] = e1;
			else
				query->table->pdata[i] = e0;
		}
		else
		{
			tmp = paillier_plaintext_from_ui((unsigned long int) raw_hash[i]);
			query->table->pdata[i] = paillier_enc(0, query->pub, tmp,
																						paillier_get_rand_devurandom);
			paillier_freeplaintext(tmp);
		}

		if( !(i % 3) || i + 1 == query->table->len )
		{
			status("\rencrypting query table ... %6d / %-6d", i + 1, query->table->len);
			fflush(stdout);
		}
	}

	status("\n");
}

void
set_id( enc_query_t* q )
{
	srand(time(0));
	q->id = ((uint64_t) rand())<<32 | rand();
}

int
main( int argc, char** argv )
{
	int* raw_hash;
	qkey_t* qkey;
	enc_query_t* query;

	pseudo_rand_init();

	/* create query and query private key */
	query = (enc_query_t*) malloc(sizeof(enc_query_t));
	set_id(query);
	query->table = g_ptr_array_new();
	qkey = (qkey_t*) malloc(sizeof(qkey_t));
	qkey->id = query->id;

	/* parse args, setting globals and stuff in query and qkey */
	parse_args(argc, argv, query, qkey);

	/* set performance / correctness parameters */
	set_params(query);

	/* create raw query table */
	raw_hash = create_table(qkey->words, query->table->len);

	/* make key pair */
	status("generating key pair (move mouse to add to entropy pool) ...\n");
 	paillier_keygen(keybits, &query->pub, &qkey->prv, paillier_get_rand_devrandom);

	/* encrypt query table */
	build_query(raw_hash, query);

	/* save encrypted query and query private key */
	write_query(qoutfile, query);
	write_qkey(koutfile, qkey);

	return 0;
}
