/*
	privss - A private stream searching toolkit.

	Original version copyright (C) 2006 SRI International.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
	General Public License for more details.

  Later changes copyright John Bethencourt, also released under the
  GPLv2 or later.
*/

#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif
#include <unistd.h>
#include <gmp.h>
#include <iml.h>

#include <stdio.h>
#include <stdint.h>
#include <glib.h>
#include <paillier.h>
#include "privss.h"
#include "pseudorand.h"

#define NONSINGMIN 30

char* usage =
"Usage: privss-recon [OPTION ...] QUERYFILE RESFILE PRIVKEY\n"
"\n"
"Reconstruct the files which matched the query QUERYFILE, based on the results\n"
"RESFILE and the private key PRIVKEY.\n"
"\n"
"Mandatory arguments to long options are mandatory for short options too.\n\n"
" -h, --help               print this message\n\n"
" -v, --version            print version information\n\n"
" -q, --quiet              don't print any status messages\n\n"
" -b, --verbose            print extensive status messages\n\n"
" -o, --output-dir DIR     produce reconstructed files in DIR; default is\n"
"                          current directory\n\n"
"";

char* output_dir = ".";
char* query_file = 0;
char* res_file   = 0;
char* key_file   = 0;

/* parses args and fills in global options */

void
parse_args( int argc, char** argv )
{
	int i;

	for( i = 1; i < argc; i++ )
		if(      !strcmp(argv[i], "-h") || !strcmp(argv[i], "--help") )
		{
			printf("%s", usage);
			exit(0);
		}
		else if( !strcmp(argv[i], "-v") || !strcmp(argv[i], "--version") )
		{
			printf(PRIVSS_VERSION_MSG, "-recon");
			exit(0);
		}
		else if( !strcmp(argv[i], "-q") || !strcmp(argv[i], "--quiet") )
		{
			set_verbosity(VERBOSITY_QUIET);
		}
		else if( !strcmp(argv[i], "-b") || !strcmp(argv[i], "--verbose") )
		{
			set_verbosity(VERBOSITY_LOUD);
		}
		else if( !strcmp(argv[i], "-o") || !strcmp(argv[i], "--output-dir") )
		{
			if( ++i >= argc )
				die(usage);
			else
				output_dir = argv[i];
		}
		else if( !query_file )
		{
			query_file = argv[i];
		}
		else if( !res_file )
		{
			res_file = argv[i];
		}
		else if( !key_file )
		{
			key_file = argv[i];
		}
		else
			die(usage);

	if( !query_file || !res_file || !key_file )
		die(usage);

	if( output_dir[strlen(output_dir)-1] == '/' )
		output_dir[strlen(output_dir)-1] = 0;
}

void
print_dec_prog( int dec_so_far, int to_dec )
{
	if( !(dec_so_far % 10) || dec_so_far == to_dec )
	{
		status("\rdecrypting results ... %7d / %-7d", dec_so_far, to_dec);
		fflush(stdout);
	}
}

GPtrArray* /* paillier_plaintext_t* */
decrypt_buffer( GPtrArray* b, /* paillier_ciphertext_t* */
								paillier_pubkey_t* pub,
								paillier_prvkey_t* prv,
								int* dec_so_far,
								int  to_dec )
{
	GPtrArray* r;
	int i;

	r = g_ptr_array_new();
	g_ptr_array_set_size(r, b->len);

	for( i = 0; i < r->len; i++ )
	{
		r->pdata[i] = paillier_dec(0, pub, prv, b->pdata[i]);
		print_dec_prog(++*dec_so_far, to_dec);
	}

	return r;
}

void
read_and_decrypt( enc_query_t** query,
									qkey_t** qkey,
									uint32_t* num_searched,
									GPtrArray** i_buf, /* paillier_plaintext_t* */
									GPtrArray** c_buf, /* paillier_plaintext_t* */
									GPtrArray** l_buf, /* paillier_plaintext_t* */
									GPtrArray** d_buf  /* paillier_plaintext_t* (2D, block-major) */ )
{
	enc_results_t* res;
	int to_dec;
	int dec_so_far;

	*query = read_query(query_file);
	*qkey = read_qkey(key_file, (*query)->pub);
	res = read_results(res_file, *query, 0);
	if( (*query)->id != res->id || (*query)->id != (*qkey)->id )
		die("%s, %s, %s not from a single search\n", query_file, res_file, key_file);

	*num_searched = res->num_searched;

	status_verb("%d documents were searched ...\n", res->num_searched);

	dec_so_far = 0;
	to_dec = res->c_buf->len + res->l_buf->len + res->d_buf->len + res->i_buf->len;
	*i_buf = decrypt_buffer(res->i_buf, (*query)->pub, (*qkey)->prv, &dec_so_far, to_dec);
	*c_buf = decrypt_buffer(res->c_buf, (*query)->pub, (*qkey)->prv, &dec_so_far, to_dec);
	*l_buf = decrypt_buffer(res->l_buf, (*query)->pub, (*qkey)->prv, &dec_so_far, to_dec);
	*d_buf = decrypt_buffer(res->d_buf, (*query)->pub, (*qkey)->prv, &dec_so_far, to_dec);
	status("\n");

	free_results(res);
}

int
check_bf_index( enc_query_t* query, int i, GPtrArray* i_buf, mpz_t mask )
{
	int j0, j1;
	mpz_t t;
	int r;

	j0 = i / query->sfactor;
	j1 = i % query->sfactor;

	if( !mpz_sgn(((paillier_plaintext_t*) i_buf->pdata[j0])->m) )
		return 0;

	/* get appropriate chunk */
	mpz_init_set(t, ((paillier_plaintext_t*) i_buf->pdata[j0])->m);

	/* shift right so the block we want is in the least sig bits */
	mpz_tdiv_q_2exp(t, t, j1 * ((query->pub->bits - 16) / query->sfactor));

	/* clear out other crud */
	mpz_and(t, t, mask);

	r = !!mpz_sgn(t);
	mpz_clear(t);

	return r;
}

GArray*
check_bf( enc_query_t* query, uint32_t num_searched, GPtrArray* i_buf )
{
	GHashTable* h;
	GArray* indices;
	mpz_t mask;
	int i, j;

	indices = g_array_new(0, 1, sizeof(uint32_t));

	h = g_hash_table_new(0, 0);

	/* this is used in each call to check_bf_index, we just compute it
		 once here for efficiency */
	mpz_init_set_ui(mask, 1);
	mpz_mul_2exp(mask, mask, (query->pub->bits - 16) / query->sfactor);
	mpz_sub_ui(mask, mask, 1);

	status_verb("bloom filter yields indices (starting with 1, [k] indicates padding): ");

	/* check the filter */
	for( i = 1; i <= num_searched; i++ )
	{
		int not_match = 0;

		for( j = 0; j < query->num_bf_hashes && !not_match; j++ )
 			if( !check_bf_index(query,
													pseudo_rand2(i, j) %
													(i_buf->len * query->sfactor), i_buf, mask) )
 				not_match = 1;
		
 		if( !not_match )
		{
			g_array_append_val(indices, i);
			g_hash_table_insert(h, GINT_TO_POINTER(i), GINT_TO_POINTER(1));
			status_verb(" %d", i);
		}
	}

	/* pad things out with phony indices to prevent a singular matrix */
	while( indices->len < NONSINGMIN )
	{
		i = 1;
		while( g_hash_table_lookup(h, GINT_TO_POINTER(i)) )
			i++;
		g_array_append_val(indices, i);
		g_hash_table_insert(h, GINT_TO_POINTER(i), GINT_TO_POINTER(1));
		status_verb(" [%d]", i);
	}

	status_verb("\n");

	g_hash_table_destroy(h);
	mpz_clear(mask);
	
	return indices;
}

/* for redirecting stdout to stop IML from shitting all over the terminal */
int saved_stdout_fd = -1;

void
ditch_stdout()
{
#ifdef HAVE_FCNTL_H
#ifdef HAVE_DUP2
	int new_fd;
	fflush(stdout);
  new_fd = open("/dev/null", O_WRONLY, 0666);
  saved_stdout_fd = dup(fileno(stdout));
  dup2(new_fd, fileno(stdout));
  close(new_fd);
#endif
#endif
}

void
unditch_stdout()
{
#ifdef HAVE_FCNTL_H
#ifdef HAVE_DUP2
	fflush(stdout);
  dup2(saved_stdout_fd, fileno(stdout));
  close(saved_stdout_fd);
  saved_stdout_fd = -1;
#endif
#endif
}

void
compute_matrix_a1( enc_query_t* query, GArray* indices, mpz_t** N, mpz_t** D )
{
	long int* A;
	int i, j;
	mpz_t* I;

	/* set up A */
	A = (long int*) malloc(sizeof(long int) * indices->len * indices->len);
	for( i = 0; i < indices->len; i++ )
		for( j = 0; j < indices->len; j++ )
			A[i * indices->len + j] =
				!(pseudo_rand2(g_array_index(indices, uint32_t, j), i) % 2);

	/* set up I */
	I = (mpz_t*) malloc(sizeof(mpz_t) * indices->len * indices->len);
	for( i = 0; i < indices->len; i++ )
		for( j = 0; j < indices->len; j++ )
			mpz_init_set_ui(I[i*indices->len + j], i == j);

	/* set up space for results */
	*N = (mpz_t*) malloc(sizeof(mpz_t) * indices->len * indices->len);
	*D = (mpz_t*) malloc(sizeof(mpz_t));
	for( i = 0; i < indices->len; i++ )
		for( j = 0; j < indices->len; j++ )
			mpz_init((*N)[i*indices->len + j]);
	mpz_init(**D);

	/* invoke IML */
	status("solving linear system of %d variables ...\n", indices->len);
	ditch_stdout();
	nonsingSolvMM(RightSolu, indices->len, indices->len, A, I, *N, **D);
	unditch_stdout();

	if( !mpz_invert(**D, **D, query->pub->n) )
		die("error: %s has no inverse mod %s!\n",
				mpz_get_str(0, 10, **D), mpz_get_str(0, 10, query->pub->n));

	/* mod these beasts */
	for( i = 0; i < indices->len; i++ )
		for( j = 0; j < indices->len; j++ )
			mpz_mod((*N)[i*indices->len + j], (*N)[i*indices->len + j], query->pub->n);
	mpz_mod(**D, **D, query->pub->n);

	/* cleanup */
	for( i = 0; i < indices->len; i++ )
		for( j = 0; j < indices->len; j++ )
			mpz_clear(I[i*indices->len + j]);
	free(I);
	free(A);
}

void
compute_cvals( enc_query_t* query, int num_matches,
							 mpz_t* invan, mpz_t invad,
							 GPtrArray* c_buf, mpz_t** cvals )
{
	int i, j;

	status_verb("cvals:");

	*cvals = (mpz_t*) malloc(sizeof(mpz_t) * num_matches);
	for( i = 0; i < num_matches; i++ )
		mpz_init_set_ui((*cvals)[i], 0);

	for( i = 0; i < num_matches; i++ )
		for( j = 0; j < num_matches; j++ )
			mpz_addmul((*cvals)[i], invan[i*num_matches + j],
								 ((paillier_plaintext_t*) c_buf->pdata[j])->m);

	for( i = 0; i < num_matches; i++ )
	{
		mpz_mul((*cvals)[i], (*cvals)[i], invad);
		mpz_mod((*cvals)[i], (*cvals)[i], query->pub->n);
		status_verb(" %d", mpz_get_ui((*cvals)[i]));
	}

	status_verb("\n");
}

void
cut_matrix( int row, int col, int oldlen, mpz_t* mat )
{
	int i;
	int inew, jnew;
	int iold, jold;

	iold = jold = 0;
	for( inew = 0; inew < oldlen - 1; inew++ )
		for( jnew = 0; jnew < oldlen - 1; jnew++ )
		{
			if( jold == col )
			{
				jold++;
				if( jold >= oldlen )
				{
					jold = 0;
					iold++;
				}
			}

			if( iold == row )
				iold++;

			mpz_set(mat[inew * (oldlen-1) + jnew],
							mat[iold * oldlen + jold]);

			jold++;
			if( jold >= oldlen )
			{
				jold = 0;
				iold++;
			}
		}

	for( i = (oldlen - 1) * (oldlen - 1); i < oldlen * oldlen; i++ )
		mpz_clear(mat[i]);
}

void
filter_cvals( enc_query_t* query, GArray* indices, mpz_t* invan, mpz_t* cvals )
{
	int i, j;

	for( i = 0; i < indices->len; i++ )
		if( !mpz_sgn(cvals[i]) )
		{
			if( indices->len > NONSINGMIN )
			{
				/* fix indices */
				g_array_remove_index(indices, i);

				/* fix cvals */
				/* note subtlety of array bound: indices->len just decremented */
				for( j = i; j < indices->len; j++ )
					mpz_swap(cvals[j], cvals[j+1]);
				mpz_clear(cvals[indices->len]);

				/* fix matrix */
				cut_matrix(indices->len, i, indices->len + 1, invan);

				i--;
			}
			else
				mpz_set_ui(cvals[i], 1);
		}
}

GArray*
compute_lvals( enc_query_t* query, GArray* indices, mpz_t* invan, mpz_t invad,
							 mpz_t* cvals, GPtrArray* l_buf, uint32_t* total_len )
{
	GArray* lengths;
	mpz_t s, t;
	int i, j;

	lengths = g_array_new(0, 1, sizeof(uint32_t));
	mpz_init(s);
	mpz_init(t);

	status_verb("lengths (index, length):");

	for( i = 0; i < indices->len; i++ )
	{
		uint32_t l;

		mpz_set_ui(t, 0);
		for( j = 0; j < indices->len; j++ )
		{
			mpz_addmul(t, invan[i * indices->len + j],
								 ((paillier_plaintext_t*) l_buf->pdata[j])->m);
			mpz_mod(t, t, query->pub->n);
		}

		mpz_invert(s, cvals[i], query->pub->n);
		mpz_mul(t, t, s);
		mpz_mul(t, t, invad);
		mpz_mod(t, t, query->pub->n);

		l = mpz_get_ui(t);
		if( l == 0 )
			break;
		else
		{
			g_array_append_val(lengths, l);
			status_verb(" (%d,%d)", g_array_index(indices, uint32_t, i), l);
		}
	}

	status_verb("\n");

	mpz_clear(s);
	mpz_clear(t);

	*total_len = 0;
	for( i = 0; i < lengths->len; i++ )
		*total_len += g_array_index(lengths, uint32_t, i);

	return lengths;
}

char* c_buf_overflow_msg =
"\n"
"\"C Buffer\" Overflow! (c buffer length %d, needed at least %d)\n"
"\n"
"Too many documents matched the query for the results to be correctly\n"
"recovered. Try giving the --max-results option to privss-qcon to allow\n"
"more space. Also, false positives in the query hash table contribute\n"
"to the number of matching documents at this stage in recovery and can\n"
"thus cause overflow. These false positives can be reduced by specifying\n"
"a larger hash table with the --table-size or --fp-rate options to\n"
"privss-qcon.\n"
"\n";

char* l_buf_overflow_msg =
"\n"
"\"L Buffer\" Overflow! (l buffer length %d, needed at least %d)\n"
"\n"
"Too many documents matched the query for the results to be correctly\n"
"recovered. Try giving the --max-results option to privss-qcon to allow\n"
"more space. Also, false positives in the query hash table contribute\n"
"to the number of matching documents at this stage in recovery and can\n"
"thus cause overflow. These false positives can be reduced by specifying\n"
"a larger hash table with the --table-size or --fp-rate options to\n"
"privss-qcon.\n"
"\n";

char* d_buf_overflow_msg =
"\n"
"\"D Buffer\" Overflow! (d buffer length %d, needed at least %d)\n"
"\n"
"The matching documents were too large to be correctly recovered. Try\n"
"giving the --max-result-size option to privss-qcon to allow more space.\n"
"Also, false positives in the query hash table contribute to the number\n"
"of matching documents at this stage in recovery and can thus cause\n"
"overflow. These false positives can be reduced by specifying a larger\n"
"hash table with the --table-size or --fp-rate options to privss-qcon.\n"
"\n";

void
check_overflow( int a, int b, char* msg )
{
	if( a > b )
		die(msg, b, a);
}

void
compute_metadata( /* input */
								  enc_query_t* query,
									uint32_t num_searched,
									GPtrArray* i_buf, /* paillier_plaintext_t* */
									GPtrArray* c_buf, /* paillier_plaintext_t* */
									GPtrArray* l_buf, /* paillier_plaintext_t* */
									GPtrArray* d_buf, /* paillier_plaintext_t* (2D, block-major) */
									/* output */
									GArray**   indices, /* uint32_t */
									mpz_t**    cvals,
									GArray**   lengths, /* uint32_t */
									uint32_t*  total_len )
{
	mpz_t* invan;
	mpz_t* invad;
	int i;

	/* reconstruct matching indices from bf */
	*indices = check_bf(query, num_searched, i_buf);
	check_overflow((*indices)->len, c_buf->len, c_buf_overflow_msg);
	status_verb("total of %d <= c_buf->len (%d)\n", (*indices)->len, c_buf->len);

	/* compute inverse of A_1 as two integer parts */
	compute_matrix_a1(query, *indices, &invan, &invad);

	/* compute c values */
	compute_cvals(query, (*indices)->len, invan, *invad, c_buf, cvals);
	
	/* filter out bf false positives, adjust matrix appropriately */
	filter_cvals(query, *indices, invan, *cvals);
	check_overflow((*indices)->len, l_buf->len, l_buf_overflow_msg);
	status_verb("filtered to total of %d <= l_buf->len (%d)\n", (*indices)->len, l_buf->len);

	/* compute doc lengths in blocks */
	*lengths = compute_lvals(query, *indices, invan, *invad, *cvals, l_buf, total_len);
	check_overflow(*total_len, d_buf->len / query->block_size, d_buf_overflow_msg);
	status_verb("total of %d <= d_buf->len (%d)\n", *total_len, d_buf->len / query->block_size);
	status_verb("(totaled across %d docs)\n", (*lengths)->len);

	/* cleanup */
	for( i = 0; i < (*indices)->len * (*indices)->len; i++ )
		mpz_clear(invan[i]);
	free(invan);
	mpz_clear(*invad);
	free(invad);
}

void
compute_matrix_a2( enc_query_t* query, GArray* indices,
									 GArray* lengths, uint32_t total_len, /* both in blocks */
									 mpz_t** N, mpz_t** D )
{
	int i, j;
	uint32_t docnum, docblock;
	long int* A;
	mpz_t* I;

	/* set up A */
	A = (long int*) malloc(sizeof(long int) * total_len * total_len);
	for( i = 0; i < total_len; i++ )
	{
		docnum = 0;
		docblock = 0;
		for( j = 0; j < total_len; j++ )
		{
			A[i * total_len + j] =
				!(pseudo_rand3(g_array_index(indices, uint32_t, docnum), docblock, i) % 2);

			docblock++;
			if( docnum < lengths->len &&
          docblock >= g_array_index(lengths, uint32_t, docnum) )
			{
				docnum++;
				docblock = 0;
			}
		}
	}

	/* set up I */
	I = (mpz_t*) malloc(sizeof(mpz_t) * total_len * total_len);
	for( i = 0; i < total_len; i++ )
		for( j = 0; j < total_len; j++ )
			mpz_init_set_ui(I[i*total_len + j], i == j);

	/* set up space for results */
	*N = (mpz_t*) malloc(sizeof(mpz_t) * total_len * total_len);
	*D = (mpz_t*) malloc(sizeof(mpz_t));
	for( i = 0; i < total_len; i++ )
		for( j = 0; j < total_len; j++ )
			mpz_init((*N)[i*total_len + j]);
	mpz_init(**D);

	/* invoke IML */
	status("solving linear system of %d variables ...\n", total_len);
	ditch_stdout();
	nonsingSolvMM(RightSolu, total_len, total_len, A, I, *N, **D);
	unditch_stdout();

	if( !mpz_invert(**D, **D, query->pub->n) )
		die("error: %s has no inverse mod %s!\n",
				mpz_get_str(0, 10, **D), mpz_get_str(0, 10, query->pub->n));

	/* mod these beasts */
	for( i = 0; i < total_len; i++ )
		for( j = 0; j < total_len; j++ )
			mpz_mod((*N)[i*total_len + j], (*N)[i*total_len + j], query->pub->n);
	mpz_mod(**D, **D, query->pub->n);

	/* cleanup */
	for( i = 0; i < total_len; i++ )
		for( j = 0; j < total_len; j++ )
			mpz_clear(I[i*total_len + j]);
	free(I);
	free(A);
}

doc_t*
compute_doc( enc_query_t* query, mpz_t* inva2n, mpz_t inva2d,
						 uint32_t offset  /* in blocks */, uint32_t total_blocks,
						 mpz_t cval, uint32_t numblocks, GPtrArray* d_buf )
{
	doc_t* doc;
	GPtrArray* pts;
	paillier_plaintext_t* pt;
	int i, j, k;
	mpz_t cinv;

	mpz_init(cinv);
	mpz_invert(cinv, cval, query->pub->n);

	pts = g_ptr_array_new();
	for( i = offset; i < offset + numblocks; i++ )
		/* compute this block */
		for( k = 0; k < query->block_size; k++ )
		{
			pt = paillier_plaintext_from_ui(0);
			for( j = 0; j < total_blocks; j++ )
			{
				mpz_addmul(pt->m, inva2n[i * total_blocks + j],
									 ((paillier_plaintext_t*) d_buf->pdata[j * query->block_size + k])->m);
				mpz_mod(pt->m, pt->m, query->pub->n);
			}

			mpz_mul(pt->m, pt->m, cinv);
			mpz_mul(pt->m, pt->m, inva2d);
			mpz_mod(pt->m, pt->m, query->pub->n);

			g_ptr_array_add(pts, pt);
		}

	doc = deserialize_doc(query, pts);

	for( i = 0; i < pts->len; i++ )
		paillier_freeplaintext(pts->pdata[i]);
	g_ptr_array_free(pts, 0);

	mpz_clear(cinv);

	return doc;
}

GPtrArray*
compute_docs( enc_query_t* query, GArray* indices, mpz_t* cvals,
							GArray* lengths, uint32_t total_len, /* both in blocks */
							GPtrArray* d_buf )
{
	mpz_t* inva2n;
	mpz_t* inva2d;
	GPtrArray* docs; /* doc_t* */
	uint32_t offset; /* in blocks */
	int i;

	docs = g_ptr_array_new();
	if( total_len == 0 )
		/* don't go into the matrix stuff if there is nothing to compute */
		return docs;
  else if( total_len < NONSINGMIN )
    /* ensure A_2 is reasonable minimum size to prevent singularity */
    total_len = NONSINGMIN;

	/* compute inverse of A_2 as two integer parts */
	compute_matrix_a2(query, indices, lengths, total_len, &inva2n, &inva2d);

	/* compute each document */
	offset = 0;
	for( i = 0; i < lengths->len; offset += g_array_index(lengths, uint32_t, i++) )
		g_ptr_array_add(docs, compute_doc(query, inva2n, *inva2d, offset, total_len, cvals[i],
                                      g_array_index(lengths, uint32_t, i), d_buf));
  
	/* cleanup */
	for( i = 0; i < total_len * total_len; i++ )
		mpz_clear(inva2n[i]);
	free(inva2n);
	mpz_clear(*inva2d);
	free(inva2d);

	return docs;
}

gint
strcmpp( gconstpointer a, gconstpointer b )
{
	return strcmp(a, b);
}

void
filter_docs( qkey_t* qkey, GPtrArray* docs )
{
	int i;
	GSList* l;
	GSList* lp;

	for( i = 0; i < docs->len; i++ )
	{
		if( !((doc_t*)docs->pdata[i])->words )
		{
			/* doc does not include keywords, do not filter */
			status_verb("keeping keyword-less doc %d (%s)\n",
									i, ((doc_t*)docs->pdata[i])->filename);
			continue;
		}

		l = qkey->words;
		while( l )
		{
			/* look for this word in doc */
			if( (lp = g_slist_find_custom(((doc_t*)docs->pdata[i])->words, l->data, strcmpp)) )
				break;
			l = l->next;
		}
		if( !l )
		{
			/* remove this doc */
			status_verb("doc %s filtered due to query fp\n", ((doc_t*)docs->pdata[i])->filename);
			g_ptr_array_remove_index(docs, i--);
		}
	}
}

void
save_docs( enc_query_t* query, GPtrArray* docs )
{
	int i;

	for( i = 0; i < docs->len; i++ )
	{
		status("saving file %s ...\n", ((doc_t*)docs->pdata[i])->filename);
		write_doc(docs->pdata[i], output_dir);
	}
}

void
free_cvals( mpz_t* cvals, int k )
{
	int i;

	for( i = 0; i < k; i++ )
		mpz_clear(cvals[i]);
	free(cvals);
}

int
main( int argc, char** argv )
{
	/* starting search info */
	enc_query_t* query;
	qkey_t* qkey;
	uint32_t num_searched;
	GPtrArray* i_buf; /* paillier_plaintext_t* */
	GPtrArray* c_buf; /* paillier_plaintext_t* */
	GPtrArray* l_buf; /* paillier_plaintext_t* */
	GPtrArray* d_buf; /* paillier_plaintext_t* (2D, block-major) */

	/* computed metadata and results */
	GArray* indices; /* uint32_t */
	mpz_t* 	cvals;
	GArray* lengths; /* uint32_t (in blocks) */
	uint32_t total_len; /* (also in blocks) */
	GPtrArray* docs; /* doc_t* */

	pseudo_rand_init();

	/* parse args */
	parse_args(argc, argv);

	/* read query, results, private key and decrypt results buffers */
  read_and_decrypt(&query, &qkey, &num_searched, &i_buf, &c_buf, &l_buf, &d_buf);

	/* compute all the information we need to reconstruct the actual documents */
	compute_metadata(query, num_searched, i_buf, c_buf, l_buf, d_buf,
									 &indices, &cvals, &lengths, &total_len);

	/* cleanup */
	free_parray(i_buf);
	free_parray(c_buf);
	free_parray(l_buf);

	/* compute documents based on metadata and d_buf */
	docs = compute_docs(query, indices, cvals, lengths, total_len, d_buf);

	/* remove query table false positives */
	filter_docs(qkey, docs);
	status("%d documents matched query\n", docs->len);

	/* cleanup */
	free_cvals(cvals, indices->len);
	g_array_free(indices, 0);
	g_array_free(lengths, 0);
	free_parray(d_buf);

	/* save them in files */
	save_docs(query, docs);

	return 0;
}
