/*
	privss - A private stream searching toolkit.

	Original version copyright (C) 2006 SRI International.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
	General Public License for more details.

  Later changes copyright John Bethencourt, also released under the
  GPLv2 or later.
*/

#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include <stdio.h>
#include <stdint.h>
#include <glib.h>
#include <gmp.h>
#include <paillier.h>
#include "privss.h"
#include "pseudorand.h"

/*
	TODO

  - fix multiple document processing mode. the parsing is total shit.
*/

char* usage =
"Usage: privss-search [OPTION ...] QUERYFILE RESFILE [DOCFILE WORD [WORD ...]]\n"
"\n"
"Use QUERYFILE to run the private search algorithm. The partial results stored\n"
"in RESFILE are loaded and updated. If RESFILE does not exist, a new search\n"
"will be started.\n"
"\n"
"If DOCFILE and at least one WORD are specified, privss-search will run the\n"
"search algorithm on the document DOCFILE with the associated words, write the\n"
"current results to RESFILE, then exit. Otherwise, privss-search will read\n"
"lines of the form\n"
"\n"
" DOCFILE WORD [WORD ...]\n"
"\n"
"from stdin, process each in turn, and save the current results in RESFILE\n"
"when the end of file is reached.\n"
"\n"
"Mandatory arguments to long options are mandatory for short options too.\n\n"
" -h, --help               print this message\n\n"
" -v, --version            print version information\n\n"
" -q, --quiet              don't print status messages\n\n"
" -w, --word-file FILE     read additional words (one per line) from FILE; in\n"
"                          this case DOCFILE must be specified and it is not\n"
"                          necessary to specify any keywords on the command\n"
"                          line\n\n"
"";

char* 	query_file        = 0;
char* 	res_file   			  = 0;
char* 	single_doc_file   = 0;
char* 	word_file         = 0;
GSList* single_words      = 0;

/* parses args and fills in global options */

void
parse_args( int argc, char** argv )
{
	int i;

	for( i = 1; i < argc; i++ )
		if(      !strcmp(argv[i], "-h") || !strcmp(argv[i], "--help") )
		{
			printf("%s", usage);
			exit(0);
		}
		else if( !strcmp(argv[i], "-v") || !strcmp(argv[i], "--version") )
		{
			printf(PRIVSS_VERSION_MSG, "-search");
			exit(0);
		}
		else if( !strcmp(argv[i], "-q") || !strcmp(argv[i], "--quiet") )
		{
			set_verbosity(VERBOSITY_QUIET);
		}
		else if( !strcmp(argv[i], "-w") || !strcmp(argv[i], "--word-file") )
		{
			if( ++i >= argc )
				die(usage);
			else
				word_file = argv[i];
		}
		else if( !query_file )
		{
			query_file = argv[i];
		}
		else if( !res_file )
		{
			res_file = argv[i];
		}
		else if( !single_doc_file )
		{
			single_doc_file = strdup(argv[i]);
		}
		else
		{
			single_words = g_slist_append(single_words, strdup(argv[i]));
		}

	if( word_file )
	{
		FILE* fp;
		GString* s;
		char c;

		if( !(fp = fopen(word_file, "r")) )
			die("couldn't read file: %s\n", word_file);

		s = g_string_sized_new(5);
		while( !feof(fp) )
			if( (c = fgetc(stdin)) == '\n' )
			{
				if( *(s->str) )
					single_words = g_slist_append(single_words, s->str);
				g_string_free(s, 0);
				s = g_string_sized_new(32);
			}
			else
				s = g_string_append_c(s, c);

		if( *(s->str) )
			single_words = g_slist_append(single_words, s->str);
		g_string_free(s, 0);

		fclose(fp);
	}

	if( !query_file || !res_file || (single_doc_file && !single_words) )
		die(usage);
}

paillier_ciphertext_t*
compute_cval( enc_query_t* query, GSList* words )
{
	paillier_ciphertext_t* ec;

	ec = paillier_create_enc_zero();
	while( words )
	{
		paillier_mul(query->pub, ec, ec,
								 query->table->pdata
								 [pseudo_rands((char*) words->data) % query->table->len]);
		words = words->next;
	}

	return ec;
}

paillier_ciphertext_t*
compute_lval( enc_query_t* query, int numblock, paillier_ciphertext_t* ec )
{
	paillier_ciphertext_t* el;
	paillier_plaintext_t* pt;

	el = paillier_create_enc_zero();
	pt = paillier_plaintext_from_ui(numblock);
	paillier_exp(query->pub, el, ec, pt);
	paillier_freeplaintext(pt);

	return el;
}

/* paillier_ciphertext_t* (2D, block-major) */
GPtrArray*
compute_cdval( enc_query_t* query, GPtrArray* docpts, int reallen,
							 paillier_ciphertext_t* ec )
{
	GPtrArray* ecd;
	int i;

	ecd = g_ptr_array_new();

	for( i = 0; i < docpts->len; i++ )
	{
		g_ptr_array_add(ecd, paillier_create_enc_zero());
		if( i < reallen ) /* speed hack */
			paillier_exp(query->pub, ecd->pdata[i], ec, docpts->pdata[i]);
	}

	return ecd;
}

void
update_c_buf( enc_query_t* query, enc_results_t* results,
							paillier_ciphertext_t* ec )
{
	int i;

	for( i = 0; i < results->c_buf->len; i++ )
		if( !(pseudo_rand2(results->num_searched, i) % 2) )
			paillier_mul(query->pub, results->c_buf->pdata[i],
									 results->c_buf->pdata[i], ec);
}

void
update_l_buf( enc_query_t* query, enc_results_t* results,
							paillier_ciphertext_t* el )
{
	int i;

	for( i = 0; i < results->l_buf->len; i++ )
		if( !(pseudo_rand2(results->num_searched, i) % 2) )
			paillier_mul(query->pub, results->l_buf->pdata[i],
									 results->l_buf->pdata[i], el);
}

void
update_d_buf( enc_query_t* query, enc_results_t* results,
							/* paillier_ciphertext_t* (2D, block-major) */
							GPtrArray* ecd )
{
	int i, j, k;

	for( i = 0; i < ecd->len / query->block_size; i++ )
		for( j = 0; j < results->d_buf->len / query->block_size; j++ )
			if( !(pseudo_rand3(results->num_searched, i, j) % 2) )
				for( k = 0; k < query->block_size; k++ )
					paillier_mul(query->pub,
											 results->d_buf->pdata[j * query->block_size + k],
											 results->d_buf->pdata[j * query->block_size + k],
											 ecd->pdata[i * query->block_size + k]);
}

void
update_i_buf_index( enc_query_t* query, enc_results_t* results,
										int i, paillier_ciphertext_t* ec )
{
	paillier_ciphertext_t* ec_shifted;
	paillier_plaintext_t* shifter;
	int j0, j1;
	uint8_t* buf;
	int b;

	j0 = i / query->sfactor;
	j1 = i % query->sfactor;

	/* bits in each block */
	b = (query->pub->bits - 16) / query->sfactor;

	/* set shifter = 2^{j1 b} and compute ec_shifted = ec^shifter */
	buf = malloc(PAILLIER_BITS_TO_BYTES(j1 * b + 1));
	memset(buf, 0, PAILLIER_BITS_TO_BYTES(j1 * b + 1));
	buf[0] = 1<<((j1 * b) % 8);
	shifter = paillier_plaintext_from_bytes(buf, PAILLIER_BITS_TO_BYTES(j1 * b + 1));
	free(buf);

	ec_shifted = paillier_create_enc_zero();
	paillier_exp(query->pub, ec_shifted, ec, shifter);

	paillier_mul(query->pub, results->i_buf->pdata[j0],
							 results->i_buf->pdata[j0], ec_shifted);

	paillier_freeplaintext(shifter);
	paillier_freeciphertext(ec_shifted);
}

void
update_i_buf( enc_query_t* query, enc_results_t* results,
							paillier_ciphertext_t* ec )
{
	int i;

	for( i = 0; i < query->num_bf_hashes; i++ )
		update_i_buf_index
			(query, results,
			 pseudo_rand2(results->num_searched, i) %
			 (results->i_buf->len * query->sfactor), ec);
}

/*
	Get a string of all the chars until the next newline or eof,
	skipping initial whitespace. Returns zero if eof is reached.
*/
GString*
next_line()
{
	GString* s;
	int c;

	/* skip initial whitespace */
	c = fgetc(stdin);
	while( c != EOF && isspace(c) )
		c = fgetc(stdin);
	if( c != EOF )
		ungetc(c, stdin);

	/* get line */
	s = g_string_sized_new(32);
	while( !feof(stdin) )
	{
		c = fgetc(stdin);
		if( c == EOF || c == '\n' )
			break;
		else
			s = g_string_append_c(s, c);
	}

	if( s->len == 0 )
	{
		g_string_free(s, 1);
		return 0;
	}

	return s;
}

int
next_document( char** doc_file, GSList** words )
{
	GString* s;
	char** a;
	char** ap;

	s = next_line();
	if( s == 0 )
		return 0;
	
	*words = 0;

	ap = a = g_strsplit(s->str, " ", 0);
	while( *ap )
		if( strlen(*ap) > 0 )
			*words = g_slist_append(*words, strdup(*(ap++)));
		else
			ap++;
	g_strfreev(a);

	g_string_free(s, 1);

	if( g_slist_length(*words) < 2 )
		die(usage);
	else
	{
		*doc_file = (*words)->data;
		*words = g_slist_remove(*words, *doc_file);
	}

	return 1;
}

void
process_doc( enc_query_t* query, enc_results_t* results,
						 char* doc_file, GSList* words )
{
	paillier_ciphertext_t* ec;
	paillier_ciphertext_t* el;
	doc_t* doc;
	GPtrArray* ecd;    /* paillier_ciphertext_t* */
	GPtrArray* docpts; /* paillier_plaintext_t* */
	int reallen;

	/* get started */
	results->num_searched++;
	status("processing document %d: %s ...\n", results->num_searched, doc_file);

	/* create document and serialize it into plaintext blocks */
	doc = read_doc(doc_file, words);
	docpts = serialize_doc(query, doc, &reallen);
	free_doc(doc);

	/* compute stuff */
	ec  = compute_cval(query, words);
	el  = compute_lval(query, docpts->len / query->block_size, ec);
	ecd = compute_cdval(query, docpts, reallen, ec);

	/* update buffers */
	update_i_buf(query, results, ec);
	update_c_buf(query, results, ec);
	update_l_buf(query, results, el);
	update_d_buf(query, results, ecd);

	/* cleanup */
	free_parray(docpts);
	paillier_freeciphertext(ec);
	paillier_freeciphertext(el);
	free_carray(ecd);
}

int
main( int argc, char** argv )
{
	enc_query_t* query;
	enc_results_t* results;
	char* doc_file;
	GSList* words;
	
	pseudo_rand_init();

	/* parse args and load query */
	parse_args(argc, argv);
	query = read_query(query_file);

	/* load current buffers (or initialize new buffers) */
	results = read_results(res_file, query, 1);

	if( single_doc_file )
		/* process one document from command line */
		process_doc(query, results, single_doc_file, single_words);
	else
		/* process each document from stdin */
		while( next_document(&doc_file, &words) )
		{
			process_doc(query, results, doc_file, words);
			free(doc_file);
			free_slist_items(words);
		}

	/* write buffers */
	write_results(res_file, query->pub, results);

	/* this is really unnecessary, but I want to shut up valgrind */
	free_query(query);
	free_results(results);

	return 0;
}
